@@ -1,5 +1,5 @@
 // config.c (part of mintty)
-// Copyright 2008-13 Andy Koppe
+// Copyright 2008-13 Andy Koppe, 2015-2016 Thomas Wolff
 // Based on code from PuTTY-0.60 by Simon Tatham and team.
 // Licensed under the terms of the GNU General Public License v3 or later.
 
@@ -12,8 +12,12 @@
 #include <termios.h>
 #include <sys/cygwin.h>
 
+
 #define dont_debug_config
 
+#define dont_support_blurred
+
+
 #if CYGWIN_VERSION_API_MINOR >= 222
 static wstring rc_filename = 0;
 #else
@@ -30,6 +34,7 @@ const config default_cfg = {
   .search_current_colour = 0x0099DD,
   .cursor_colour = 0xBFBFBF,
   .transparency = 0,
+  .blurred = false,
   .opaque_when_focused = false,
   .cursor_type = CUR_LINE,
   .cursor_blinks = true,
@@ -52,8 +57,11 @@ const config default_cfg = {
   .zoom_font_with_window = true,
   .alt_fn_shortcuts = true,
   .ctrl_shift_shortcuts = false,
-  .break_string = "",
-  .pause_string = "",
+  .key_prtscreen = "",	// VK_SNAPSHOT
+  .key_pause = "",	// VK_PAUSE
+  .key_break = "",	// VK_CANCEL
+  .key_menu = "",	// VK_APPS
+  .key_scrlock = "",	// VK_SCROLL
   // Mouse
   .copy_on_select = true,
   .copy_as_rtf = true,
@@ -74,31 +82,34 @@ const config default_cfg = {
   .search_bar = "",
   // Terminal
   .term = "xterm",
-  .answerback = "",
+  .answerback = L"",
   .bell_sound = false,
-  .bell_type = 0,
+  .bell_type = 1,
+  .bell_file = L"",
   .bell_freq = 0,
   .bell_len = 400,
   .bell_flash = false,
   .bell_taskbar = true,
-  .printer = "",
+  .printer = L"",
   .confirm_exit = true,
   // Command line
-  .class = "",
+  .class = L"",
   .hold = HOLD_START,
   .exit_write = false,
-  .exit_title = "",
-  .icon = "",
-  .log = "",
+  .exit_title = L"",
+  .icon = L"",
+  .log = L"",
   .utmp = false,
-  .title = "",
+  .title = L"",
   .daemonize = true,
+  .daemonize_always = false,
   // "Hidden"
-  .app_id = "",
-  .app_name = "",
-  .app_launch_cmd = "",
+  .app_id = L"",
+  .app_name = L"",
+  .app_launch_cmd = L"",
   .col_spacing = 0,
   .row_spacing = 0,
+  .padding = 1,
   .word_chars = "",
   .word_chars_excl = "",
   .use_system_colours = false,
@@ -129,7 +140,7 @@ static config file_cfg;
 typedef enum {
   OPT_BOOL, OPT_MOD, OPT_TRANS, OPT_CURSOR, OPT_FONTSMOOTH,
   OPT_MIDDLECLICK, OPT_RIGHTCLICK, OPT_SCROLLBAR, OPT_WINDOW, OPT_HOLD,
-  OPT_INT, OPT_COLOUR, OPT_STRING,
+  OPT_INT, OPT_COLOUR, OPT_STRING, OPT_WSTRING,
   OPT_LEGACY = 16
 } opt_type;
 
@@ -150,6 +161,9 @@ options[] = {
   {"SearchCurrentColour", OPT_COLOUR, offcfg(search_current_colour)},
   {"CursorColour", OPT_COLOUR, offcfg(cursor_colour)},
   {"Transparency", OPT_TRANS, offcfg(transparency)},
+#ifdef support_blurred
+  {"Blur", OPT_BOOL, offcfg(blurred)},
+#endif
   {"OpaqueWhenFocused", OPT_BOOL, offcfg(opaque_when_focused)},
   {"CursorType", OPT_CURSOR, offcfg(cursor_type)},
   {"CursorBlinks", OPT_BOOL, offcfg(cursor_blinks)},
@@ -176,8 +190,13 @@ options[] = {
   {"ZoomFontWithWindow", OPT_BOOL, offcfg(zoom_font_with_window)},
   {"AltFnShortcuts", OPT_BOOL, offcfg(alt_fn_shortcuts)},
   {"CtrlShiftShortcuts", OPT_BOOL, offcfg(ctrl_shift_shortcuts)},
-  {"Break", OPT_STRING, offcfg(break_string)},
-  {"Pause", OPT_STRING, offcfg(pause_string)},
+  {"Key_PrintScreen", OPT_STRING, offcfg(key_prtscreen)},
+  {"Key_Pause", OPT_STRING, offcfg(key_pause)},
+  {"Key_Break", OPT_STRING, offcfg(key_break)},
+  {"Key_Menu", OPT_STRING, offcfg(key_menu)},
+  {"Key_ScrollLock", OPT_STRING, offcfg(key_scrlock)},
+  {"Break", OPT_STRING | OPT_LEGACY, offcfg(key_break)},
+  {"Pause", OPT_STRING | OPT_LEGACY, offcfg(key_pause)},
 
   // Mouse
   {"CopyOnSelect", OPT_BOOL, offcfg(copy_on_select)},
@@ -201,36 +220,39 @@ options[] = {
 
   // Terminal
   {"Term", OPT_STRING, offcfg(term)},
-  {"Answerback", OPT_STRING, offcfg(answerback)},
+  {"Answerback", OPT_WSTRING, offcfg(answerback)},
   {"BellSound", OPT_BOOL, offcfg(bell_sound)},
   {"BellType", OPT_INT, offcfg(bell_type)},
+  {"BellFile", OPT_WSTRING, offcfg(bell_file)},
   {"BellFreq", OPT_INT, offcfg(bell_freq)},
   {"BellLen", OPT_INT, offcfg(bell_len)},
   {"BellFlash", OPT_BOOL, offcfg(bell_flash)},
   {"BellTaskbar", OPT_BOOL, offcfg(bell_taskbar)},
-  {"Printer", OPT_STRING, offcfg(printer)},
+  {"Printer", OPT_WSTRING, offcfg(printer)},
   {"ConfirmExit", OPT_BOOL, offcfg(confirm_exit)},
 
   // Command line
-  {"Class", OPT_STRING, offcfg(class)},
+  {"Class", OPT_WSTRING, offcfg(class)},
   {"Hold", OPT_HOLD, offcfg(hold)},
   {"Daemonize", OPT_BOOL, offcfg(daemonize)},
+  {"DaemonizeAlways", OPT_BOOL, offcfg(daemonize_always)},
   {"ExitWrite", OPT_BOOL, offcfg(exit_write)},
-  {"ExitTitle", OPT_STRING, offcfg(exit_title)},
-  {"Icon", OPT_STRING, offcfg(icon)},
-  {"Log", OPT_STRING, offcfg(log)},
-  {"Title", OPT_STRING, offcfg(title)},
+  {"ExitTitle", OPT_WSTRING, offcfg(exit_title)},
+  {"Icon", OPT_WSTRING, offcfg(icon)},
+  {"Log", OPT_WSTRING, offcfg(log)},
+  {"Title", OPT_WSTRING, offcfg(title)},
   {"Utmp", OPT_BOOL, offcfg(utmp)},
   {"Window", OPT_WINDOW, offcfg(window)},
   {"X", OPT_INT, offcfg(x)},
   {"Y", OPT_INT, offcfg(y)},
 
   // "Hidden"
-  {"AppID", OPT_STRING, offcfg(app_id)},
-  {"AppName", OPT_STRING, offcfg(app_name)},
-  {"AppLaunchCmd", OPT_STRING, offcfg(app_launch_cmd)},
+  {"AppID", OPT_WSTRING, offcfg(app_id)},
+  {"AppName", OPT_WSTRING, offcfg(app_name)},
+  {"AppLaunchCmd", OPT_WSTRING, offcfg(app_launch_cmd)},
   {"ColSpacing", OPT_INT, offcfg(col_spacing)},
   {"RowSpacing", OPT_INT, offcfg(row_spacing)},
+  {"Padding", OPT_INT, offcfg(padding)},
   {"WordChars", OPT_STRING, offcfg(word_chars)},
   {"WordCharsExcl", OPT_STRING, offcfg(word_chars_excl)},
   {"IMECursorColour", OPT_COLOUR, offcfg(ime_cursor_colour)},
@@ -441,7 +463,7 @@ parse_colour(string s, colour *cp)
 }
 
 static int
-set_option(string name, string val_str)
+set_option(string name, string val_str, bool from_file)
 {
   int i = find_option(name);
   if (i < 0)
@@ -454,6 +476,16 @@ set_option(string name, string val_str)
     when OPT_STRING:
       strset(val_p, val_str);
       return i;
+    when OPT_WSTRING: {
+      wchar * ws;
+      if (from_file)
+        ws = cs__utforansitowcs(val_str);
+      else
+        ws = cs__mbstowcs(val_str);
+      wstrset(val_p, ws);
+      free(ws);
+      return i;
+    }
     when OPT_INT: {
       char *val_end;
       int val = strtol(val_str, &val_end, 0);
@@ -490,7 +522,7 @@ set_option(string name, string val_str)
 }
 
 static int
-parse_option(string option)
+parse_option(string option, bool from_file)
 {
   const char *eq = strchr(option, '=');
   if (!eq) {
@@ -511,7 +543,7 @@ parse_option(string option)
   while (isspace((uchar)*val))
     val++;
 
-  return set_option(name, val);
+  return set_option(name, val, from_file);
 }
 
 static void
@@ -526,13 +558,13 @@ check_arg_option(int i)
 void
 set_arg_option(string name, string val)
 {
-  check_arg_option(set_option(name, val));
+  check_arg_option(set_option(name, val, false));
 }
 
 void
 parse_arg_option(string option)
 {
-  check_arg_option(parse_option(option));
+  check_arg_option(parse_option(option, false));
 }
 
 void
@@ -561,7 +593,7 @@ load_config(string filename, bool to_save)
     while (fgets(line, sizeof line, file)) {
       line[strcspn(line, "\r\n")] = 0;  /* trim newline */
       if (line[0] != '#' && line[0] != '\0') {
-        int i = parse_option(line);
+        int i = parse_option(line, true);
         if (to_save && i >= 0)
           remember_file_option(i);
       }
@@ -586,6 +618,8 @@ copy_config(config *dst_p, const config *src_p)
       switch (type) {
         when OPT_STRING:
           strset(dst_val_p, *(string *)src_val_p);
+        when OPT_WSTRING:
+          wstrset(dst_val_p, *(wstring *)src_val_p);
         when OPT_INT or OPT_COLOUR:
           *(int *)dst_val_p = *(int *)src_val_p;
         otherwise:
@@ -658,6 +692,11 @@ save_config(void)
         switch (type) {
           when OPT_STRING:
             fprintf(file, "%s", *(string *)val_p);
+          when OPT_WSTRING: {
+            char * s = cs__wcstoutf(*(wstring *)val_p);
+            fprintf(file, "%s", s);
+            free(s);
+          }
           when OPT_INT:
             fprintf(file, "%i", *(int *)val_p);
           when OPT_COLOUR: {
@@ -704,6 +743,8 @@ apply_config(bool save)
     switch (type) {
       when OPT_STRING:
         changed = strcmp(*(string *)val_p, *(string *)new_val_p);
+      when OPT_WSTRING:
+        changed = wcscmp(*(wstring *)val_p, *(wstring *)new_val_p);
       when OPT_INT or OPT_COLOUR:
         changed = (*(int *)val_p != *(int *)new_val_p);
       otherwise:
@@ -762,21 +803,30 @@ current_size_handler(control *unused(ctrl), int event)
 static void
 printerbox_handler(control *ctrl, int event)
 {
-  const string NONE = "None (printing disabled)";
-  string printer = new_cfg.printer;
+  const wstring NONE = L"◇ None (printing disabled) ◇";  // ♢◇
+  const wstring CFG_NONE = L"";
+  const wstring DEFAULT = L"◆ Default printer ◆";  // ♦◆
+  const wstring CFG_DEFAULT = L"*";
+  wstring printer = new_cfg.printer;
   if (event == EVENT_REFRESH) {
     dlg_listbox_clear(ctrl);
-    dlg_listbox_add(ctrl, NONE);
+    dlg_listbox_add_w(ctrl, NONE);
+    dlg_listbox_add_w(ctrl, DEFAULT);
     uint num = printer_start_enum();
     for (uint i = 0; i < num; i++)
-      dlg_listbox_add(ctrl, printer_get_name(i));
+      dlg_listbox_add_w(ctrl, (wchar *)printer_get_name(i));
     printer_finish_enum();
-    dlg_editbox_set(ctrl, *printer ? printer : NONE);
+    if (*printer == '*')
+      dlg_editbox_set_w(ctrl, DEFAULT);
+    else
+      dlg_editbox_set_w(ctrl, *printer ? printer : NONE);
   }
   else if (event == EVENT_VALCHANGE || event == EVENT_SELCHANGE) {
-    dlg_editbox_get(ctrl, &printer);
-    if (!strcmp(printer, NONE))
-      strset(&printer, "");
+    dlg_editbox_get_w(ctrl, &printer);
+    if (!wcscmp(printer, NONE))
+      wstrset(&printer, CFG_NONE);
+    else if (!wcscmp(printer, DEFAULT))
+      wstrset(&printer, CFG_DEFAULT);
     new_cfg.printer = printer;
   }
 }
@@ -874,6 +924,57 @@ term_handler(control *ctrl, int event)
   }
 }
 
+//  1 -> 0x00000000 MB_OK              Default Beep
+//  2 -> 0x00000010 MB_ICONSTOP        Critical Stop
+//  3 -> 0x00000020 MB_ICONQUESTION    Question
+//  4 -> 0x00000030 MB_ICONEXCLAMATION Exclamation
+//  5 -> 0x00000040 MB_ICONASTERISK    Asterisk
+// -1 -> 0xFFFFFFFF                    Simple Beep
+static string beeps[] = {
+  "simple",
+  "none",
+  "default",
+  "Critical Stop",
+  "Question",
+  "Exclamation",
+  "Asterisk",
+};
+
+static void
+bell_handler(control *ctrl, int event)
+{
+  switch (event) {
+    when EVENT_REFRESH:
+      dlg_listbox_clear(ctrl);
+      int sel = -1;
+      for (uint i = 0; i < lengthof(beeps); i++) {
+        dlg_listbox_add(ctrl, beeps[i]);
+        if ((int)i == new_cfg.bell_type + 1)
+          sel = i;
+      }
+      if (sel >= 0)
+        dlg_editbox_set(ctrl, beeps[sel]);
+    when EVENT_VALCHANGE or EVENT_SELCHANGE: {
+      char * beep = malloc(0);
+      dlg_editbox_get(ctrl, (string *)&beep);
+      for (uint i = 0; i < lengthof(beeps); i++) {
+        if (!strcmp(beep, beeps[i])) {
+          new_cfg.bell_type = i - 1;
+          break;
+        }
+      }
+      free(beep);
+    }
+  }
+}
+
+static void
+bell_tester(control *unused(ctrl), int event)
+{
+  if (event == EVENT_ACTION)
+    win_bell(&new_cfg);
+}
+
 void
 setup_config_box(controlbox * b)
 {
@@ -924,10 +1025,22 @@ setup_config_box(controlbox * b)
     with_glass ? "Gla&ss" : null, TR_GLASS,
     null
   );
+#ifdef support_blurred
+  ctrl_columns(s, 2, with_glass ? 80 : 75, with_glass ? 20 : 25);
+  ctrl_checkbox(
+    s, "Opa&que when focused",
+    dlg_stdcheckbox_handler, &new_cfg.opaque_when_focused
+  )->column = 0;
+  ctrl_checkbox(
+    s, "&Blur",
+    dlg_stdcheckbox_handler, &new_cfg.blurred
+  )->column = 1;
+#else
   ctrl_checkbox(
     s, "Opa&que when focused",
     dlg_stdcheckbox_handler, &new_cfg.opaque_when_focused
   );
+#endif
 
   s = ctrl_new_set(b, "Looks", "Cursor");
   ctrl_radiobuttons(
@@ -1139,10 +1252,10 @@ setup_config_box(controlbox * b)
     s, "&Answerback", 100, dlg_stdstringbox_handler, &new_cfg.answerback
   )->column = 1;
 
-  s = ctrl_new_set(b, "Terminal", "Bell");
-  ctrl_columns(s, 3, 25, 25, 50);
-  ctrl_checkbox(
-    s, "&Sound", dlg_stdcheckbox_handler, &new_cfg.bell_sound
+  s = ctrl_new_set(b, "Terminal", "Bell sound (overridden by BellFreq or BellFile)");
+  ctrl_columns(s, 3, 36, 19, 45);
+  ctrl_combobox(
+    s, null, 100, bell_handler, 0
   )->column = 0;
   ctrl_checkbox(
     s, "&Flash", dlg_stdcheckbox_handler, &new_cfg.bell_flash
@@ -1150,11 +1263,34 @@ setup_config_box(controlbox * b)
   ctrl_checkbox(
     s, "&Highlight in taskbar", dlg_stdcheckbox_handler, &new_cfg.bell_taskbar
   )->column = 2;
+  ctrl_columns(s, 1, 100);  // reset column stuff so we can rearrange them
+  ctrl_columns(s, 2, 80, 20);
+#ifdef use_bellfileselection
+  ctrl_combobox(
+    s, "&Wave", 80,
+    bellfile_selector, &new_cfg.bell_file
+  )->column = 0;
+#else
+#ifdef use_belleditbox
+  ctrl_editbox(
+    s, "&Wave", 80, dlg_stdstringbox_handler, &new_cfg.bell_file
+  )->column = 0;
+#endif
+#endif
+  ctrl_pushbutton(
+    s, "&Play", bell_tester, 0
+  )->column = 1;
 
   s = ctrl_new_set(b, "Terminal", "Printer");
+#ifdef use_multi_listbox_for_printers
+  ctrl_listbox(
+    s, null, 4, 100, printerbox_handler, 0
+  );
+#else
   ctrl_combobox(
     s, null, 100, printerbox_handler, 0
   );
+#endif
 
   s = ctrl_new_set(b, "Terminal", null);
   ctrl_checkbox(
