@@ -21,12 +21,13 @@ void win_schedule_update(void);
 void win_text(int x, int y, wchar *text, int len, cattr attr, int lattr);
 void win_update_mouse(void);
 void win_capture_mouse(void);
-void win_bell(void);
+void win_bell(config *);
 
 void win_set_title(char *);
 void win_save_title(void);
 void win_restore_title(void);
-void win_prefix_title(const char *);
+void win_prefix_title(const wstring);
+void win_unprefix_title(const wstring);
 void win_copy_title(void);
 
 colour win_get_colour(colour_i);
