@@ -1,6 +1,7 @@
 // winctrls.c (part of mintty)
-// Copyright 2008-11 Andy Koppe
+// Copyright 2008-11 Andy Koppe, 2015-2016 Thomas Wolff
 // Adapted from code from PuTTY-0.60 by Simon Tatham and team.
+// (corresponds to putty:windows/winctrls.c)
 // Licensed under the terms of the GNU General Public License v3 or later.
 
 #include "winctrls.h"
@@ -322,9 +323,15 @@ button(ctrlpos * cp, char *btext, int bid, int defbtn)
   if (defbtn && cp->wnd)
     SendMessage(cp->wnd, DM_SETDEFID, bid, 0);
 
-  doctl(cp, r, "BUTTON",
-        BS_NOTIFY | WS_CHILD | WS_VISIBLE | WS_TABSTOP |
-        (defbtn ? BS_DEFPUSHBUTTON : 0) | BS_PUSHBUTTON, 0, btext, bid);
+  HWND but = 
+    doctl(cp, r, "BUTTON",
+          BS_NOTIFY | WS_CHILD | WS_VISIBLE | WS_TABSTOP |
+          (defbtn ? BS_DEFPUSHBUTTON : 0) | BS_PUSHBUTTON, 0, btext, bid);
+  // this is a special hack until a generic solution is crafted 
+  // for Unicode labels and maybe other fancy stuff
+  if (!strcmp(btext, "&Play")) {
+    SendMessageW(but, WM_SETTEXT, 0, (LPARAM)L"► &Play");
+  }
 
   cp->ypos += PUSHBTNHEIGHT + GAPBETWEEN;
 }
@@ -506,8 +513,9 @@ winctrl_cleanup(winctrls *wc)
   winctrl *c = wc->first;
   while (c) {
     winctrl *next = c->next;
-    if (c->ctrl)
+    if (c->ctrl) {
       c->ctrl->plat_ctrl = null;
+    }
     free(c->data);
     free(c);
     c = next;
@@ -970,12 +978,12 @@ winctrl_handle_command(UINT msg, WPARAM wParam, LPARAM lParam)
             when CBN_SELCHANGE: {
               int index = SendDlgItemMessage(
                             dlg.wnd, c->base_id + 1, CB_GETCURSEL, 0, 0);
-              int len = SendDlgItemMessage(
+              int wlen = SendDlgItemMessageW(
                           dlg.wnd, c->base_id + 1, CB_GETLBTEXTLEN, index, 0);
-              char text[len + 1];
-              SendDlgItemMessage(
-                dlg.wnd, c->base_id + 1, CB_GETLBTEXT, index, (LPARAM) text);
-              SetDlgItemText(dlg.wnd, c->base_id + 1, text);
+              wchar wtext[wlen + 1];
+              SendDlgItemMessageW(
+                dlg.wnd, c->base_id + 1, CB_GETLBTEXT, index, (LPARAM) wtext);
+              SetDlgItemTextW(dlg.wnd, c->base_id + 1, wtext);
               ctrl->handler(ctrl, EVENT_SELCHANGE);
             }
             when CBN_EDITCHANGE:
@@ -1069,11 +1077,44 @@ dlg_editbox_set(control *ctrl, string text)
   SetDlgItemText(dlg.wnd, c->base_id + 1, text);
 }
 
+void
+dlg_editbox_set_w(control *ctrl, wstring text)
+{
+  winctrl *c = ctrl->plat_ctrl;
+  assert(c &&
+         (c->ctrl->type == CTRL_EDITBOX
+         ||c->ctrl->type == CTRL_LISTBOX));
+  if (c->ctrl->type != CTRL_LISTBOX) {
+    SetDlgItemTextW(dlg.wnd, c->base_id + 1, text);
+  }
+  else {
+    HWND wnd = GetDlgItem(dlg.wnd, c->base_id + 1);
+    int len = wcslen(text);
+    wchar * buf = newn(wchar, len + 1);
+    int n = SendMessageW(wnd, LB_GETCOUNT, 0, (LPARAM)0);
+    for (int i = 0; i < n; i++) {
+      int ilen = SendMessageW(wnd, LB_GETTEXTLEN, i, (LPARAM)0);
+      if (ilen > len) {
+        buf = renewn(buf, ilen + 1);
+        len = ilen;
+      }
+      SendMessageW(wnd, LB_GETTEXT, i, (LPARAM)buf);
+      if (wcscmp(buf, text) == 0) {
+        SendMessageW(wnd, LB_SETCURSEL, i, (LPARAM)0);
+        break;
+      }
+    }
+    free(buf);
+  }
+}
+
 void
 dlg_editbox_get(control *ctrl, string *text_p)
 {
   winctrl *c = ctrl->plat_ctrl;
-  assert(c && c->ctrl->type == CTRL_EDITBOX);
+  assert(c &&
+         (c->ctrl->type == CTRL_EDITBOX
+         ||c->ctrl->type == CTRL_LISTBOX));
   HWND wnd = GetDlgItem(dlg.wnd, c->base_id + 1);
   int size = GetWindowTextLength(wnd) + 1;
   char *text = (char *)*text_p;
@@ -1082,6 +1123,37 @@ dlg_editbox_get(control *ctrl, string *text_p)
   *text_p = text;
 }
 
+void
+dlg_editbox_get_w(control *ctrl, wstring *text_p)
+{
+  winctrl *c = ctrl->plat_ctrl;
+  assert(c &&
+         (c->ctrl->type == CTRL_EDITBOX
+         ||c->ctrl->type == CTRL_LISTBOX));
+  HWND wnd = GetDlgItem(dlg.wnd, c->base_id + 1);
+  wchar *text = (wchar *)*text_p;
+  if (c->ctrl->type != CTRL_LISTBOX) {
+    // handle single-line editbox (with optional popup list)
+    int size = GetWindowTextLengthW(wnd) + 1;
+    text = renewn(text, size);
+    // In the popup editbox (combobox), 
+    // Windows goofs up non-ANSI characters here unless the 
+    // WM_COMMAND dialog callback function winctrl_handle_command above 
+    // (case CBN_SELCHANGE) also uses the Unicode function versions 
+    // SendDlgItemMessageW and SetDlgItemTextW
+    GetWindowTextW(wnd, text, size);
+    //GetDlgItemTextW(dlg.wnd, c->base_id + 1, text, size);  // same
+  }
+  else {
+    // handle multi-line listbox
+    int n = SendMessageW(wnd, LB_GETCURSEL, 0, (LPARAM)0);
+    int len = SendMessageW(wnd, LB_GETTEXTLEN, n, (LPARAM)0);
+    text = renewn(text, len + 1);
+    SendMessageW(wnd, LB_GETTEXT, n, (LPARAM)text);
+  }
+  *text_p = text;
+}
+
 /* The `listbox' functions also apply to combo boxes. */
 void
 dlg_listbox_clear(control *ctrl)
@@ -1111,6 +1183,20 @@ dlg_listbox_add(control *ctrl, string text)
   SendDlgItemMessage(dlg.wnd, c->base_id + 1, msg, 0, (LPARAM) text);
 }
 
+void
+dlg_listbox_add_w(control *ctrl, wstring text)
+{
+  winctrl *c = ctrl->plat_ctrl;
+  int msg;
+  assert(c &&
+         (c->ctrl->type == CTRL_LISTBOX ||
+          (c->ctrl->type == CTRL_EDITBOX &&
+           c->ctrl->editbox.has_list)));
+  msg = (c->ctrl->type == CTRL_LISTBOX &&
+         c->ctrl->listbox.height != 0 ? LB_ADDSTRING : CB_ADDSTRING);
+  SendDlgItemMessageW(dlg.wnd, c->base_id + 1, msg, 0, (LPARAM) text);
+}
+
 void
 dlg_fontsel_set(control *ctrl, font_spec *fs)
 {
